import re
from base64 import b32encode, b32decode
from jupyterhub.handlers import BaseHandler
from jupyterhub.auth import Authenticator
from jupyterhub.auth import LocalAuthenticator
from jupyterhub.utils import url_path_join
from tornado import gen, web
from traitlets import Unicode, List
from ast import literal_eval


def safeinput_encode(input_str):
    """
    :param input_str: string
    :return: b32encoded utf-8 string with stripped padding equals
    """
    encoded_str = str(b32encode(bytes(input_str, 'utf-8')), 'utf-8')
    return encoded_str.replace('=', '')


def safeinput_decode(input_str):
    """
    :param input_str: expects a b32encoded utf-8 string
    :return: a decoded utf-8 string
    """
    # Encoder removed "=" padding to satisfy validate_input
    # Pad with "="" according to:
    # https://tools.ietf.org/html/rfc3548 :
    # (1) the final quantum of encoding input is an integral multiple of 40
    # bits; here, the final unit of encoded output will be an integral
    # multiple of 8 characters with no "=" padding.
    if len(input_str) % 8 != 0:
        padlen = 8 - (len(input_str) % 8)
        padding = "".join('=' for i in range(padlen))
        decode_str = "{}{}".format(input_str, padding)
    else:
        decode_str = input_str

    return str(b32decode(bytes(decode_str, 'utf-8')), 'utf-8')


def extract_headers(request, headers):
    user_data = {}
    for i, header in enumerate(headers):
        value = request.headers.get(header, "")
        if value:
            try:
                user_data[header] = value
            except KeyError:
                pass
    return user_data


class PartialBaseURLHandler(BaseHandler):
    """
    Fix against /base_url requests are not redirected to /base_url/home
    """
    @web.authenticated
    @gen.coroutine
    def get(self):
        self.redirect(url_path_join(self.hub.server.base_url, 'home'))


class RemoteUserLogoutHandler(BaseHandler):

    @gen.coroutine
    def get(self):
        user = self.get_current_user()
        if user:
            self.clear_login_cookie()
        self.redirect(self.hub.server.base_url)


class RemoteUserLoginHandler(BaseHandler):

    @gen.coroutine
    def prepare(self):
        """ login user """
        if self.get_current_user() is not None:
            self.log.info("User: {} is already authenticated"
                          .format(self.get_current_user(), self.get_current_user().name))
            self.redirect(url_path_join(self.hub.server.base_url, 'home'))
        else:
            user_data = extract_headers(self.request, self.authenticator.auth_headers)
            if 'Remote-User' not in user_data:
                raise web.HTTPError(401, "You are not Authenticated to do this")
            yield self.login_user(user_data)

            argument = self.get_argument("next", None, True)
            if argument is not None:
                self.redirect(argument)
            else:
                self.redirect(url_path_join(self.hub.server.base_url, 'home'))


class DataHandler(BaseHandler):
    """
    If the request is properly authenticated, check for a valid HTTP header,
    Excepts a string structure that can be interpreted by the ast module.
    If valid the passed information is appended to the authenticated user's state data
    dictionary where the header name is used as the key value.
    """

    @web.authenticated
    @gen.coroutine
    def post(self):
        user_data = extract_headers(self.request,
                                    self.authenticator.data_headers)
        if not user_data:
            raise web.HTTPError(403, "No valid data header was received")

        self.log.debug("Prepared user_data dict: {}".format(user_data))
        user = self.get_current_user()
        for k, d in user_data.items():
            # Try to parse the passed information into a valid dtype
            try:
                evaled_data = literal_eval(d)
            except ValueError as err:
                msg = "Failed to interpret the data header"
                self.log.error("User: {} - {}-{}-{}".format(
                    user, d, msg, err))
                raise web.HTTPError(403, msg)

            self.log.info("User: {}-{} Accepted data header: {}".format(
                user, user.name, evaled_data
            ))

            if not hasattr(user, 'data'):
                user.data = {}
            user.data[k] = evaled_data
        self.redirect(url_path_join(self.hub.server.base_url, 'home'))


class RemoteUserAuthenticator(Authenticator):
    """
    Accept the authenticated user name from the Remote-User HTTP header.
    """
    header_name = Unicode(
        default_value='Remote-User',
        config=True,
        help="""HTTP header to inspect for the authenticated username.""")

    def get_handlers(self, app):
        return [
            (r'/login', RemoteUserLoginHandler),
            (r'/logout', RemoteUserLogoutHandler)
        ]

    @gen.coroutine
    def authenticate(self, *args):
        raise NotImplementedError()


class RemoteUserLocalAuthenticator(LocalAuthenticator):
    """
    Accept the authenticated user name from the Remote-User HTTP header.
    Derived from LocalAuthenticator for use of features such as adding
    local accounts through the admin interface.
    """
    header_name = Unicode(
        default_value='Remote-User',
        config=True,
        help="""HTTP header to inspect for the authenticated username."""
    )

    def get_handlers(self, app):
        return [
            (r'/login', RemoteUserLoginHandler),
            (r'/logout', RemoteUserLogoutHandler)
        ]

    @gen.coroutine
    def authenticate(self, *args):
        raise NotImplementedError()


class DataRemoteUserAuthenticator(RemoteUserAuthenticator):
    """
    An Authenticator that supports dynamic headerF information
    """

    auth_headers = List(
        default_value=['Remote-User'],
        config=True,
        help="""List of allowed HTTP headers to get from user data"""
    )

    data_headers = List(
        default_value=[],
        config=True,
        help="""List of allowed data headers"""
    )

    # These paths are an extension of the prefix base url e.g. /dag/hub
    def get_handlers(self, app):
        return [
            (r'/login', RemoteUserLoginHandler),
            (r'/logout', RemoteUserLogoutHandler),
            (r'/data', DataHandler),
        ]

    @gen.coroutine
    def authenticate(self, handler, data):
        if 'Remote-User' not in data:
            self.log.info("A Remote-User header is required")
            return None

        # Login
        real_name = data['Remote-User'].lower()
        # Make it alphanumeric
        pattern = re.compile('[\W_]+', re.UNICODE)
        real_name = pattern.sub('', real_name)
        encoded_name = safeinput_encode(real_name)

        user = {
            'name': encoded_name,
            'auth_state': {
                'real_name': real_name
            }
        }
        self.log.info("Authenticated: {} - Login".format(user))
        return user

    @gen.coroutine
    def pre_spawn_start(self, user, spawner):
        """Pass upstream_token to spawner via environment variable"""
        auth_state = yield user.get_auth_state()
        if not auth_state:
            # auth_state not enabled
            return

        if isinstance(auth_state, dict) and 'real_name' in auth_state:
            user.real_name = auth_state['real_name']
            self.log.debug("Pre-Spawn: {} set user real_name {}".format(
                user, user.real_name))



import json
from json import JSONDecodeError
from re import search
from tornado import gen, web
from tornado.escape import json_decode
from jupyterhub.handlers import BaseHandler
from jupyterhub.utils import url_path_join
from traitlets import CRegExp
from traitlets.config import LoggingConfigurable


class HeaderLoginHandler(BaseHandler):
    """
    Class that is used to handle whether the user is authenticated or not
    """

    @gen.coroutine
    def prepare(self):
        """ Checks whether the user is authenticated, if so
         the user is redirected to / hub.server.base_url / home """
        if self.get_current_user():
            self.log.info("User: {} is already authenticated"
                          .format(self.get_current_user(), self.get_current_user().name))

            argument = self.get_argument("next", None, True)
            if argument:
                self.redirect(argument)
            else:
                self.redirect(url_path_join(self.hub.server.base_url, 'home'))
        else:
            # You need to authenticate first
            headers = self.request.headers
            # Authenticate user
            user = yield self.login_user(headers)
            if not user:
                raise web.HTTPError(401, "The specified authentication method failed "
                                    "to authenticate you. Please contact the system "
                                    "adminsitrator for more information")

            argument = self.get_argument("next", None, True)
            if argument:
                self.redirect(argument)
            else:
                self.redirect(url_path_join(self.hub.server.base_url, 'home'))


class UserDataHandler(BaseHandler):
    """
    Handles the post requests where an authenticated user want's to set an
    internal user attribute with a JSON body.
    """

    @web.authenticated
    @gen.coroutine
    def post(self):
        user = self.get_current_user()
        self.log.debug("UserDataHandler - Request: {}, "
                       "Body: {}".format(self.request, self.request.body))
        data = None
        try:
            data = json_decode(self.request.body)
        except JSONDecodeError as err:
            self.log.error("UserDataHandler - Failed to json decode: {}, err: {}".format(
                data, err
            ))
            raise web.HTTPError(500, "Failed to parse user data input")

        if not data:
            self.log.debug("UserDataHandler - no json data was received: {}".format(
                self.request.json
            ))
            raise web.HTTPError(403, "No data was recieved that can be used"
                                " to set an attribute")
        self.log.debug("UserDataHandler - Received: {} as a user json data post".format(
            data
        ))

        if not isinstance(data, dict):
            self.log.error("UserDataHandler - invalid internal json post structure, "
                           "expects: {}, recieved: {}".format(dict, type(data)))
            raise web.HTTPError(403, "An invalid data type was recieved")

        valid_attributes = self.authenticator.user_external_allow_attributes
        if not valid_attributes:
            raise web.HTTPError(500, "No attributes were enabled to be defined via "
                                "user data, Please contact an administrator "
                                "to resolve this")

        for valid_attr in valid_attributes:
            data_val = data.get(valid_attr, '')
            if data_val:
                try:
                    setattr(user, valid_attr, data_val)
                except AttributeError as err:
                    self.log.error("UserDataHandler - Failed to set attribute: {} "
                                   "to value: {}, err: {}".format(
                                       "self.spawner.user." + valid_attr, data_val, err
                                   ))
            else:
                self.log.debug("UserDataHandler - {} was not found "
                               "recieved data: {}".format(valid_attr, data))


class Parser(LoggingConfigurable):

    def parse(self, data):
        return data


class RegexUsernameParser(Parser):

    username_extract_regex = CRegExp(
        default_value=None,
        allow_none=False,
        help="""Regex used to extract the jupyterhub username
    from the Remote - User header"""
    ).tag(config=True)

    def parse(self, data):
        if not data:
            self.log.error("RegexUsernameParser - Didn't "
                           "receive any input missing data: {}".format(data))
            return None

        if not isinstance(data, str):
            self.log.error("RegexUsernameParser - Incorrect type was attempted to "
                           "be parsed, requires str but "
                           "data is of type: {}".format(type(data)))
            return None

        username = data
        match = search(self.username_extract_regex,
                       username)
        if not match:
            self.log.error("RegexUsernameParser - Failed to find a valid "
                           "regex match with pattern: {} in {}".format(
                               self.username_extract_regex, username))
            return None
        groups = match.groups()
        if not groups:
            self.log.error("RegexUsernameParser - No username_extract_regex "
                           "matches found in: {}".format(username))
            return None
        if len(groups) > 1:
            self.log.error("RegexUsernameParser - username_extract_regex "
                           "More than 1 match was found in: {}".format(username))
            return None
        username = groups[0]
        self.log.debug("RegexUsernameParser - Found username_extract_regex "
                       "matched: {}".format(username))
        return username


class JSONParser(Parser):

    json_types = (str, bytes, bytearray)

    def parse(self, data):
        if not data:
            self.log.error("JSONParser - Didn't "
                           "receive any input missing data: {}".format(data))
            return None
        self.log.debug("JSONParser - Data: {}, type: {}".format(data, type(data)))
        if not isinstance(data, JSONParser.json_types):
            self.log.error("JSONParser - data: {} is of an incorrect type: {} "
                           "must be one of type: {}".format(
                               data, type(data), JSONParser.json_types
                           ))
            return None

        json_obj = json.loads(data)
        return json_obj



from tornado import gen, web
from jupyterhub.auth import Authenticator
from jupyterhub.handlers.login import LogoutHandler
from traitlets import Dict, List, Type, Instance, Unicode, default
from ._jhub_shared import HeaderLoginHandler, UserDataHandler, Parser


class HeaderAuthenticator(Authenticator):
    """
    Authenticates the user via the allowed_headers
    defined 'auth' HTTP/HTTPS header.
    """
    allowed_headers = Dict(
        default_value={'auth': 'Remote-User'},
        allow_none=False,
        help="""Dict of HTTP/HTTPS headers that the authenticator will process during login.
         The 'auth' key must be set to a header, default is Remote-User.
         Additional headers will be stored in the user's 'auth_state' session store
        """
    ).tag(config=True)

    header_parser_classes = Dict(
        default_value={'auth': Parser},
        traits={Unicode(): Type(Parser, Parser)},
        allow_none=True,
        help="""Dict of classes that should be used to parse the allowed_headers dictionary
        """
    ).tag(config=True)

    header_parsers = Dict({'auth': Instance(Parser)})

    @default('header_parsers')
    def _header_parsers_default(self):
        return {parser_key: parser_val(parent=self)
                for parser_key, parser_val in self.header_parser_classes.items()}

    spawner_shared_headers = List(
        default_value=[],
        allow_none=False,
        help="""List of headers that should be shared from the auth_state dict as environment
         variables to the spawner via the pre_spawn_start hook.
        """).tag(config=True)

    user_external_allow_attributes = List(
        default_value=[],
        traits=[Unicode()],
        allow_none=True,
        help="""List of user attributes that are allowed to be defined externally
         via a JSON post request. Submit via /user-data
        """).tag(config=True)

    def __init__(self, **kwargs):
        if 'auth' not in self.allowed_headers:
            self.log.error("HeaderAuthenticator - requires that the allowed_headers "
                           "has an 'auth' key")
            raise KeyError("Missing required 'auth' key in allowed_headers")
        super().__init__(**kwargs)

    def get_handlers(self, app):
        return [
            (r'/login', HeaderLoginHandler),
            (r'/logout', LogoutHandler),
            (r'/user-data', UserDataHandler)
        ]

    @gen.coroutine
    def authenticate(self, handler, data):
        self.log.debug("HeaderAuthenticator - Request authentication with "
                       "handler: {}, data: {} of type: {}".format(handler, data,
                                                                  type(data)))
        user_data = {}
        # Process remaining allowed_headers, save valid in user_data
        for allowed_index, allowed_value in self.allowed_headers.items():
            auth_data = data.get(allowed_value, '')
            if auth_data:
                prepared_data = None
                if allowed_index in self.header_parsers:
                    prepared_data = self.header_parsers[allowed_index].parse(auth_data)
                else:
                    prepared_data = data
                if prepared_data:
                    user_data[allowed_value] = prepared_data

        self.log.debug("HeaderAuthenticator - Prepared user_data: {} "
                       "for auth check".format(user_data))
        if self.allowed_headers['auth'] not in user_data:
            self.log.error("HeaderAuthenticator - Failed to find the 'auth' key "
                           "in the user_data dictionary, this is required "
                           "to set the authenticated user's username")
            raise web.HTTPError(401, "Authentication failed, "
                                "missing information to authenticate you with")

        user = {
            'name': user_data.pop(self.allowed_headers['auth'], None)
        }
        # Something left in user_data, put in auth_state
        if user_data:
            user.update({'auth_state': user_data})

        self.log.info("Authenticated: {} - Login".format(user))
        return user

    @gen.coroutine
    def pre_spawn_start(self, user, spawner):
        """Pass upstream_token to spawner via environment variable"""
        auth_state = yield user.get_auth_state()
        if not auth_state:
            self.log.debug("HeaderAuthenticator - pre_spawn_hook, "
                           "auth_state: {}".format(auth_state))
            # auth_state not enabled
            return

        self.log.debug("HeaderAuthenticator - pre_spawn_hook, "
                       "loaded auth_state {}".format(auth_state))
        # Share permitted headers
        if not self.spawner_shared_headers:
            self.log.debug("HeaderAuthenticator - no headers were "
                           "shared with spawner environment: {}".format(
                               self.spawner_shared_headers))
            return

        for auth_key, auth_val in auth_state.items():
            if auth_key in self.spawner_shared_headers:
                spawner.environment[auth_key] = auth_val
        self.log.debug("HeaderAuthenticator - shared auth_state headers: {} with "
                       "spawner environment: {}".format(self.spawner_shared_headers,
                                                        spawner.environment))



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Juptyer Development Team.
# Distributed under the terms of the Modified BSD License.

# -----------------------------------------------------------------------------
# Minimal Python version sanity check (from IPython/Jupyterhub)
# -----------------------------------------------------------------------------

from __future__ import print_function

import os

from setuptools import setup

pjoin = os.path.join
here = os.path.abspath(os.path.dirname(__file__))

# Get the current package version.
version_ns = {}
with open(pjoin(here, 'version.py')) as f:
    exec(f.read(), {}, version_ns)

long_description = open('README.rst').read()

setup(
    name='jhub-authenticators',
    version=version_ns['__version__'],
    description='A collection of HTTP(s) JupyterHub Header Authenticators,'
    'includes Header, Remote-User and Dummy',
    long_description=long_description,
    author="Rasmus Munk",
    author_email="munk1@live.dk",
    packages=['jhubauthenticators'],
    url="https://github.com/rasmunk/jhub-authenticators",
    license="GPLv3",
    platforms="Linux, Mac OS X",
    keywords=['Interactive', 'Interpreter', 'Shell', 'Web'],
    install_requires=[
        'jupyterhub>=0.9.2',
        'docutils>=0.14'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
